#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse metadata in specific custom NOMAD OASIS ELN schema to an NXem NXS."""

# pylint: disable=no-member,duplicate-code

import numpy as np

import flatdict as fd

import yaml

from ase.data import chemical_symbols

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_versioning \
    import NX_EM_ADEF_NAME, NX_EM_ADEF_VERSION, NX_EM_EXEC_NAME, NX_EM_EXEC_VERSION


class NxEmNomadOasisElnSchemaParser:
    """Parse eln_data.yaml dump file content generated from a NOMAD OASIS YAML.

    This parser implements a design where an instance of a specific NOMAD
    custom schema ELN template is used to fill pieces of information which
    are typically not contained in files from technology partners
    (e.g. bcf, emd, dm3, ...). Until now, this custom schema and
    the NXem application definition do not use a fully harmonized vocabulary.
    Therefore, the here hardcoded implementation is needed which maps specifically
    named pieces of information from the custom schema instance on named fields
    in an instance of NXem

    The functionalities in this ELN YAML parser do not check if the
    instantiated template yields an instance which is compliant NXapm.
    Instead, this task is handled by the generic part of the dataconverter
    during the verification of the template dictionary.
    """

    def __init__(self, file_name: str, entry_id: int):
        print(f"Extracting data from ELN file: {file_name}")
        if (file_name.rsplit('/', 1)[-1].startswith("eln_data")
                or file_name.startswith("eln_data")) and entry_id > 0:
            self.entry_id = entry_id
            self.file_name = file_name
            with open(self.file_name, "r", encoding="utf-8") as stream:
                self.yml = fd.FlatDict(yaml.safe_load(stream), delimiter=":")
        else:
            self.entry_id = 1
            self.file_name = ""
            self.yml = {}

    def parse_entry_section(self, template: dict) -> dict:
        """Copy data in entry section."""
        # print("Parsing entry...")
        trg = f"/ENTRY[entry{self.entry_id}]/"
        src = "entry"
        if isinstance(self.yml[src], fd.FlatDict):
            if (self.yml[f"{src}:attr_version"] == NX_EM_ADEF_VERSION) \
                    and (self.yml[f"{src}:definition"] == NX_EM_ADEF_NAME):
                template[f"{trg}@version"] = NX_EM_ADEF_VERSION
                template[f"{trg}definition"] = NX_EM_ADEF_NAME
                template[f"{trg}PROGRAM[program1]/program"] = NX_EM_EXEC_NAME
                template[f"{trg}PROGRAM[program1]/program/@version"] = NX_EM_EXEC_VERSION
            if ("program" in self.yml[src].keys()) \
                    and ("program__attr_version" in self.yml[src].keys()):
                template[f"{trg}PROGRAM[program2]/program"] = self.yml[f"{src}:program"]
                template[f"{trg}PROGRAM[program2]/program/@version"] \
                    = self.yml[f"{src}:program__attr_version"]

        field_names = ["experiment_identifier", "start_time", "end_time",
                       "experiment_description", "experiment_documentation"]
        for field_name in field_names:
            if field_name in self.yml[src].keys():
                template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

        return template

    def parse_user_section(self, template: dict) -> dict:
        """Copy data in user section."""
        # print("Parsing user...")
        src = "user"
        if isinstance(self.yml[src], list):
            if len(self.yml[src]) >= 1:
                user_id = 1
                for user_list in self.yml[src]:
                    char_field_names = ["name", "email", "affiliation", "address",
                                        "orcid", "orcid_platform",
                                        "telephone_number", "role",
                                        "social_media_name", "social_media_platform"]

                    trg = f"/ENTRY[entry{self.entry_id}]/USER[user{user_id}]/"
                    for field_name in char_field_names:
                        if field_name in user_list.keys():
                            template[f"{trg}{field_name}"] = user_list[f"{field_name}"]
                    user_id += 1

        return template

    def parse_sample_section(self, template: dict) -> dict:
        """Copy data in sample section."""
        # check if required fields exists and are valid
        # print("Parsing sample...")
        src = "sample"
        trg = f"/ENTRY[entry{self.entry_id}]/sample/"
        if isinstance(self.yml[src], fd.FlatDict):
            if (isinstance(self.yml[f"{src}:atom_types"], list)) \
                    and (len(self.yml[src + ":atom_types"]) >= 1):
                atom_types_are_valid = True
                for symbol in self.yml[f"{src}:atom_types"]:
                    valid = isinstance(symbol, str) \
                        and (symbol in chemical_symbols) and (symbol != "X")
                    if valid is False:
                        atom_types_are_valid = False
                        break
                if atom_types_are_valid is True:
                    template[f"{trg}atom_types"] \
                        = ", ".join(list(self.yml[f"{src}:atom_types"]))

        char_req_field_names = ["method", "name", "sample_history", "preparation_date"]
        for field_name in char_req_field_names:
            if field_name in self.yml[src].keys():
                template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]
        char_opt_field_names = ["short_title", "description"]
        for field_name in char_opt_field_names:
            if field_name in self.yml[src].keys():
                template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

        float_field_names = ["thickness", "density"]
        for field_name in float_field_names:
            if (f"{field_name}:value" in self.yml[src].keys()) \
                    and (f"{field_name}:unit" in self.yml[src].keys()):
                template[f"{trg}{field_name}"] \
                    = np.float64(self.yml[f"{src}:{field_name}:value"])
                template[f"{trg}{field_name}/@units"] = self.yml[f"{src}:{field_name}:unit"]

        return template

    def parse_coordinate_system_section(self, template: dict) -> dict:
        """Define the coordinate systems to be used."""
        # ##MK::the COORDINATE_SYSTEM_SET section depends often on conventions
        # which are neither explicitly documented in instances of
        # files from technology partners nor many ELNs
        # E.g. SEM/EBSD TSL, the specific coordinate system conventions used
        # are defined in the famous "coin selector GUI window from TSL e.g."
        # but these values do not get stored in technology partner files.
        # Oxford Instruments nowadays stores coordinate systems implicitly by
        # communicating the specification of their file format (like H5OINA)
        # print("Parsing coordinate system...")
        prefix = f"/ENTRY[entry{self.entry_id}]/" \
                 f"COORDINATE_SYSTEM_SET[coordinate_system_set]/"
        # this is likely not yet matching how it should be in NeXus
        grpnm = f"{prefix}TRANSFORMATIONS[laboratory]/"
        cs_xyz = np.asarray(
            [[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], np.float64)
        cs_names = ["x", "y", "z"]
        for i in np.arange(0, 3):
            trg = f"{grpnm}AXISNAME[{cs_names[i]}]"
            template[trg] = cs_xyz[:, i]
            template[f"{trg}/@offset"] = np.asarray([0., 0., 0.], np.float64)
            template[f"{trg}/@offset_units"] = "m"
            template[f"{trg}/@depends_on"] = "."

        msg = '''
              This way of defining coordinate systems is only a small example of what
              is possible and how it can be done. More discussion among members of
              FAIRmat Areas A, B, C, and D and the EM community is needed !
              '''
        template[f"{prefix}@comment"] = msg

        return template

    def parse_instrument_header_section(self, template: dict) -> dict:
        """Copy data in instrument header section."""
        # check if required fields exists and are valid
        print("Parsing instrument header...")
        src = "em_lab"
        if isinstance(self.yml[src], fd.FlatDict):
            trg = f"/ENTRY[entry{self.entry_id}]/em_lab/"

            char_field_names = ["instrument_name", "location"]
            for field_name in char_field_names:
                if field_name in self.yml[src].keys():
                    template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

        src = "em_lab:fabrication"
        if isinstance(self.yml[src], fd.FlatDict):
            trg = f"/ENTRY[entry{self.entry_id}]/em_lab/FABRICATION[fabrication]/"
            char_field_names = ["vendor", "model", "identifier", "capabilities"]
            for field_name in char_field_names:
                if field_name in self.yml[src].keys():
                    template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

        return template

    def parse_ebeam_column_section(self, template: dict) -> dict:
        """Copy data in ebeam_column section."""
        # print("Parsing ebeam_column...")

        src = "em_lab:ebeam_column:electron_source"
        if isinstance(self.yml[src], fd.FlatDict):
            trg = f"/ENTRY[entry{self.entry_id}]/em_lab/" \
                  f"EBEAM_COLUMN[ebeam_column]/electron_source/"

            float_field_names = ["voltage"]
            for field_name in float_field_names:
                if (f"{field_name}:value" in self.yml[src].keys()) \
                        and (f"{field_name}:unit" in self.yml[src].keys()):
                    template[f"{trg}{field_name}"] \
                        = np.float64(self.yml[f"{src}:{field_name}:value"])
                    template[f"{trg}{field_name}/@units"] \
                        = self.yml[f"{src}:{field_name}:unit"]

            char_field_names = ["emitter_type"]
            for field_name in char_field_names:
                if field_name in self.yml[src].keys():
                    template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

        src = "em_lab:ebeam_column:aperture_em"
        if isinstance(self.yml[src], list):
            if len(self.yml[src]) >= 1:
                aperture_id = 1
                for aperture in self.yml[src]:
                    trg = f"/ENTRY[entry{self.entry_id}]/em_lab/" \
                          f"EBEAM_COLUMN[ebeam_column]/" \
                          f"APERTURE_EM[aperture_em{aperture_id}]/"
                    if "value" in aperture.keys():
                        template[f"{trg}value"] = np.float64(aperture["value"])
                    char_field_names = ["name", "description"]
                    for field_name in char_field_names:
                        if field_name in aperture.keys():
                            template[f"{trg}{field_name}"] = aperture[f"{field_name}"]
                    aperture_id += 1

        # the above-mentioned snippet is a blue-print for lenses also...

        # corrector_cs
        src = "em_lab:ebeam_column:aberration_correction"
        trg = f"/ENTRY[entry{self.entry_id}]/em_lab/" \
              f"EBEAM_COLUMN[ebeam_column]/aberration_correction/"
        if "applied" in self.yml[src].keys():
            template[f"{trg}applied"] = self.yml[f"{src}:applied"]

        # the inject of corrector metadata is highly instrument specific, e.g.
        # for Nion microscopes these settings are available from nionswift
        # for ThermoFisher/FEI microscopes like the Titan there is an HTML output file
        # from the Zemlin tableau analysis

        # NEW ISSUE: support parsing of corrector data (pre2 for Thilo)

        return template

    def parse_ibeam_column_section(self, template: dict) -> dict:
        """Copy data in ibeam_column section."""
        # print("Parsing ibeam_column...")
        return template

    def parse_ebeam_deflector_section(self, template: dict) -> dict:
        """Copy data in ebeam_deflector section."""
        # print("Parsing ebeam_deflector...")
        return template

    def parse_ibeam_deflector_section(self, template: dict) -> dict:
        """Copy data in ibeam_deflector section."""
        # print("Parsing ibeam_deflector...")
        return template

    def parse_optics_section(self, template: dict) -> dict:
        """Copy data in optical_system_em section."""
        # print("Parsing optics...")
        src = "em_lab:optical_system_em"
        if isinstance(self.yml[src], fd.FlatDict):
            trg = f"/ENTRY[entry{self.entry_id}]/em_lab/" \
                  f"OPTICAL_SYSTEM_EM[optical_system_em]/"

            char_field_names = ["beam_current_description"]
            for field_name in char_field_names:
                if field_name in self.yml[src].keys():
                    template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

            float_field_names = ["camera_length", "magnification", "defocus",
                                 "semi_convergence_angle", "working_distance",
                                 "beam_current"]
            for field_name in float_field_names:
                if (f"{field_name}:value" in self.yml[src].keys()) \
                        and (f"{field_name}:unit" in self.yml[src].keys()):
                    template[f"{trg}{field_name}"] \
                        = np.float64(self.yml[f"{src}:{field_name}:value"])
                    template[f"{trg}{field_name}/@units"] \
                        = self.yml[f"{src}:{field_name}:unit"]

        return template

    def parse_detector_section(self, template: dict) -> dict:
        """Copy data in detector section."""
        # print("Parsing detector...")
        src = "em_lab:detector"
        if not isinstance(self.yml[src], list):
            return template

        if len(self.yml[src]) >= 1:
            detector_id = 1
            for detector in self.yml[src]:
                if isinstance(detector, dict):
                    trg = f"/ENTRY[entry{self.entry_id}]/em_lab/" \
                          f"DETECTOR[detector{detector_id}]/"

                    char_field_names = ["local_name"]
                    for field_name in char_field_names:
                        if field_name in detector.keys():
                            template[f"{trg}{field_name}"] = detector[f"{field_name}"]
                detector_id += 1

        return template

    def parse_stage_lab_section(self, template: dict) -> dict:
        """Copy data in stage lab section."""
        # print("Parsing stage_lab...")
        src = "em_lab:stage_lab"
        if isinstance(self.yml[src], fd.FlatDict):
            trg = f"/ENTRY[entry{self.entry_id}]/em_lab/stage_lab/"

            char_field_names = ["name"]
            for field_name in char_field_names:
                if field_name in self.yml[src].keys():
                    template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]

        return template

    def report(self, template: dict) -> dict:
        """Copy data from self into template the appdef instance."""
        self.parse_entry_section(template)
        self.parse_user_section(template)
        self.parse_sample_section(template)
        self.parse_coordinate_system_section(template)
        self.parse_instrument_header_section(template)
        self.parse_ebeam_column_section(template)
        self.parse_ibeam_column_section(template)
        self.parse_ebeam_deflector_section(template)
        self.parse_ibeam_deflector_section(template)
        self.parse_optics_section(template)
        self.parse_detector_section(template)
        self.parse_stage_lab_section(template)

        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Wrapping multiple parsers for vendor files with NOMAD OASIS/ELN/YAML metadata."""

# pylint: disable=no-member,duplicate-code,too-many-nested-blocks

import flatdict as fd

import yaml

from ase.data import chemical_symbols

from pynxtools.dataconverter.readers.apm.map_concepts.apm_eln_to_nx_map \
    import NxApmElnInput, NxUserFromListOfDict

from pynxtools.dataconverter.readers.shared.map_concepts.mapping_functors \
    import variadic_path_to_specific_path, apply_modifier

from pynxtools.dataconverter.readers.apm.utils.apm_parse_composition_table \
    import parse_composition_table


class NxApmNomadOasisElnSchemaParser:  # pylint: disable=too-few-public-methods
    """Parse eln_data.yaml dump file content generated from a NOMAD OASIS YAML.

    This parser implements a design where an instance of a specific NOMAD
    custom schema ELN template is used to fill pieces of information which
    are typically not contained in files from technology partners
    (e.g. pos, epos, apt, rng, rrng, ...). Until now, this custom schema and
    the NXapm application definition do not use a fully harmonized vocabulary.
    Therefore, the here hardcoded implementation is needed which maps specifically
    named pieces of information from the custom schema instance on named fields
    in an instance of NXapm

    The functionalities in this ELN YAML parser do not check if the
    instantiated template yields an instance which is compliant NXapm.
    Instead, this task is handled by the generic part of the dataconverter
    during the verification of the template dictionary.
    """

    def __init__(self, file_name: str, entry_id: int):
        print(f"Extracting data from ELN file: {file_name}")
        if (file_name.rsplit('/', 1)[-1].startswith("eln_data")
                or file_name.startswith("eln_data")) and entry_id > 0:
            self.entry_id = entry_id
            self.file_name = file_name
            with open(self.file_name, "r", encoding="utf-8") as stream:
                self.yml = fd.FlatDict(yaml.safe_load(stream), delimiter="/")
        else:
            self.entry_id = 1
            self.file_name = ""
            self.yml = {}

    def parse_sample_composition(self, template: dict) -> dict:
        """Interpret human-readable ELN input to generate consistent composition table."""
        src = "sample/composition"
        if src in self.yml.keys():
            if isinstance(self.yml[src], list):
                dct = parse_composition_table(self.yml[src])

                prfx = f"/ENTRY[entry{self.entry_id}]/sample/" \
                       f"CHEMICAL_COMPOSITION[chemical_composition]"
                unit = "at.-%"  # the assumed default unit
                if "normalization" in dct:
                    if dct["normalization"] in ["%", "at%", "at-%", "at.-%", "ppm", "ppb"]:
                        unit = "at.-%"
                        template[f"{prfx}/normalization"] = "atom_percent"
                    elif dct["normalization"] in ["wt%", "wt-%", "wt.-%"]:
                        unit = "wt.-%"
                        template[f"{prfx}/normalization"] = "weight_percent"
                    else:
                        return template
                ion_id = 1
                for symbol in chemical_symbols[1::]:
                    # ase convention, chemical_symbols[0] == "X"
                    # to use ordinal number for indexing
                    if symbol in dct:
                        if isinstance(dct[symbol], tuple) and len(dct[symbol]) == 2:
                            trg = f"{prfx}/ION[ion{ion_id}]"
                            template[f"{trg}/name"] = symbol
                            template[f"{trg}/composition"] = dct[symbol][0]
                            template[f"{trg}/composition/@units"] = unit
                            if dct[symbol][1] is not None:
                                template[f"{trg}/composition_error"] = dct[symbol][1]
                                template[f"{trg}/composition_error/@units"] = unit
                            ion_id += 1
        return template

    def parse_user_section(self, template: dict) -> dict:
        """Copy data from user section into template."""
        src = "user"
        if src in self.yml.keys():
            if isinstance(self.yml[src], list):
                if all(isinstance(entry, dict) for entry in self.yml[src]) is True:
                    user_id = 1
                    # custom schema delivers a list of dictionaries...
                    for user_dict in self.yml[src]:
                        # ... for each of them inspect for fields mappable on NeXus
                        identifier = [self.entry_id, user_id]
                        # identifier to get instance NeXus path from variadic NeXus path
                        # try to find all quantities on the left-hand side of the mapping
                        # table and check if we can find these
                        for nx_path, modifier in NxUserFromListOfDict.items():
                            if nx_path not in ("IGNORE", "UNCLEAR"):
                                trg = variadic_path_to_specific_path(nx_path, identifier)
                                res = apply_modifier(modifier, user_dict)
                                if res is not None:
                                    template[trg] = res
                        user_id += 1
        return template

    def parse_laser_pulser_details(self, template: dict) -> dict:
        """Copy data in pulser section."""
        # additional laser-specific details only relevant when the laser was used
        src = "atom_probe/pulser/pulse_mode"
        if src in self.yml.keys():
            if self.yml[src] == "voltage":
                return template
        else:
            return template
        src = "atom_probe/pulser/laser_source"
        if src in self.yml.keys():
            if isinstance(self.yml[src], list):
                if all(isinstance(entry, dict) for entry in self.yml[src]) is True:
                    laser_id = 1
                    # custom schema delivers a list of dictionaries...
                    trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/pulser" \
                          f"/SOURCE[source{laser_id}]"
                    for laser_dict in self.yml[src]:
                        if "name" in laser_dict.keys():
                            template[f"{trg}/name"] = laser_dict["name"]
                        quantities = ["power", "pulse_energy", "wavelength"]
                        for quant in quantities:
                            if isinstance(laser_dict[quant], dict):
                                if ("value" in laser_dict[quant].keys()) \
                                        and ("unit" in laser_dict[quant].keys()):
                                    template[f"{trg}/{quant}"] \
                                        = laser_dict[quant]["value"]
                                    template[f"{trg}/{quant}/@units"] \
                                        = laser_dict[quant]["unit"]
                        laser_id += 1
        return template

    def parse_other_sections(self, template: dict) -> dict:
        """Copy data from custom schema into template."""
        for nx_path, modifier in NxApmElnInput.items():
            if nx_path not in ("IGNORE", "UNCLEAR"):
                trg = variadic_path_to_specific_path(nx_path, [self.entry_id, 1])
                res = apply_modifier(modifier, self.yml)
                if res is not None:
                    template[trg] = res
        return template

    def report(self, template: dict) -> dict:
        """Copy data from self into template the appdef instance."""
        self.parse_sample_composition(template)
        self.parse_user_section(template)
        self.parse_laser_pulser_details(template)
        self.parse_other_sections(template)
        return template



#!/usr/bin/env python
"""
Writes a NeXus HDF5 file using h5py with links to data in other HDF5 files.

This example is based on ``writer_2_1``.
"""

from pathlib import Path
import h5py
import numpy

FILE_HDF5_MASTER = "external_master.hdf5"
FILE_HDF5_ANGLES = "external_angles.hdf5"
FILE_HDF5_COUNTS = "external_counts.hdf5"

# ---------------------------

# get some data
filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
buffer = numpy.loadtxt(filename).T
tthData = buffer[0]  # float[]
countsData = numpy.asarray(buffer[1], "int32")  # int[]

# put the angle data in an external (non-NeXus) HDF5 data file
with h5py.File(FILE_HDF5_ANGLES, "w") as f:
    ds = f.create_dataset("angles", data=tthData)
    ds.attrs["units"] = "degrees"

# put the detector counts in an external HDF5 data file
# with *incomplete* NeXus structure (no NXdata group)
with h5py.File(FILE_HDF5_COUNTS, "w") as f:
    nxentry = f.create_group("entry")
    nxentry.attrs["NX_class"] = "NXentry"
    nxinstrument = nxentry.create_group("instrument")
    nxinstrument.attrs["NX_class"] = "NXinstrument"
    nxdetector = nxinstrument.create_group("detector")
    nxdetector.attrs["NX_class"] = "NXdetector"
    ds = nxdetector.create_dataset("counts", data=countsData)
    ds.attrs["units"] = "counts"
    # link the "two_theta" data stored in separate file
    local_addr = nxdetector.name + "/two_theta"
    f[local_addr] = h5py.ExternalLink(FILE_HDF5_ANGLES, "/angles")

# create a master NeXus HDF5 file
with h5py.File(FILE_HDF5_MASTER, "w") as f:
    f.attrs["default"] = "entry"
    nxentry = f.create_group("entry")
    nxentry.attrs["NX_class"] = "NXentry"
    nxentry.attrs["default"] = "data"
    nxdata = nxentry.create_group("data")
    nxdata.attrs["NX_class"] = "NXdata"

    # link in the signal data
    local_addr = "/entry/data/counts"
    external_addr = "/entry/instrument/detector/counts"
    f[local_addr] = h5py.ExternalLink(FILE_HDF5_COUNTS, external_addr)
    nxdata.attrs["signal"] = "counts"

    # link in the axes data
    local_addr = "/entry/data/two_theta"
    f[local_addr] = h5py.ExternalLink(FILE_HDF5_ANGLES, "/angles")
    nxdata.attrs["axes"] = "two_theta"
    nxdata.attrs["two_theta_indices"] = [
        0,
    ]

    local_addr = "/entry/instrument"
    f[local_addr] = h5py.ExternalLink(FILE_HDF5_COUNTS, "/entry/instrument")



import difflib
import tempfile
from pathlib import Path
from typing import List


def diff_ascii(src_file: Path, new_content: List[str], dest_file: Path) -> None:
    """`new_content` is the parsed content of `src_file` to be compared
    with the content of `dest_file`.
    """
    if dest_file.exists():
        with open(dest_file, "r") as fh:
            original_content = list(fh)
    else:
        original_content = list()

    with tempfile.TemporaryFile("w+") as fh:
        fh.writelines(new_content)
        fh.seek(0)
        new_content = list(fh)

    for line in difflib.unified_diff(
        original_content,
        new_content,
        fromfile=str(dest_file),
        tofile=str(src_file),
    ):
        print(line)



import numpy as np
import h5py
import datetime

def write_nexus_file(fname, image, md={}):
	"""
	write the image to a NeXus HDF5 data file

	Parameters
	----------
	fname : str
		name of the file (relative or absolute) to be written
	image : numpy array
		the image data
	md : dictionary
		key: value where value is something that can be written by h5py
			 (such as str, int, float, numpy array, ...)
	"""
	nexus = h5py.File(fname, "w")
	nexus.attrs["filename"] = fname
	nexus.attrs["file_time"] = datetime.datetime.now().astimezone().isoformat()
	nexus.attrs["creator"] = "write_nexus_file()"
	nexus.attrs["H5PY_VERSION"] = h5py.__version__

	# /entry
	nxentry = nexus.create_group("entry")
	nxentry.attrs["NX_class"] = "NXentry"
	nexus.attrs["default"] = nxentry.name

	# /entry/instrument
	nxinstrument = nxentry.create_group("instrument")
	nxinstrument.attrs["NX_class"] = "NXinstrument"

	# /entry/instrument/detector
	nxdetector = nxinstrument.create_group("detector")
	nxdetector.attrs["NX_class"] = "NXdetector"

	# /entry/instrument/detector/image
	ds = nxdetector.create_dataset("image", data=image, compression="gzip")
	ds.attrs["units"] = "counts"
	ds.attrs["target"] = "/entry/instrument/detector/image"

	# /entry/data
	nxdata = nxentry.create_group("data")
	nxdata.attrs["NX_class"] = "NXdata"
	nxentry.attrs["default"] = nxdata.name

	# /entry/data/data --> /entry/instrument/detector/image
	nxdata["data"] = nexus["/entry/instrument/detector/image"]
	nxdata.attrs["signal"] = "data"

	if len(md) > 0:
		# /entry/instrument/metadata (optional, for metadata)
		metadata = nxinstrument.create_group("metadata")
		metadata.attrs["NX_class"] = "NXcollection"
		for k, v in md.items():
			try:
				metadata.create_dataset(k, data=v)
			except Exception:
				metadata.create_dataset(k, data=str(v))

	nexus.close()

	
if __name__ == "__main__":
	"""demonstrate how to use this code"""
	import epics
	prefix = "13SIM1:"
	img = epics.caget(prefix+"image1:ArrayData")
	size_x = epics.caget(prefix+"cam1:ArraySizeX_RBV")
	size_y = epics.caget(prefix+"cam1:ArraySizeY_RBV")
	# edit the full image for just the binned data
	img = img[:size_x*size_y].reshape((size_x, size_y))

	extra_information = dict(
		unique_id = epics.caget(prefix+"image1:UniqueId_RBV"),
		size_x = size_x,
		size_y = size_y,
		detector_state = epics.caget(prefix+"cam1:DetectorState_RBV"),
		bitcoin_value="15000",
	)
	write_nexus_file("example.h5", img, md=extra_information)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The constraints defining if a swift display_item is assumed a NxImageAngSpace concept."""

# pylint: disable=no-member,line-too-long

# AngSpace, i.e. AngularSpace is not Reciprocal space, recall the following
# RealSpace unit category for dimension scale axes is (NX_)LENGTH
# AngSpace unit category for dimension scale axes is (NX_)ANGLE
# ReciSpace unit category for dimension scale axis is ONE_OVER_LENGTH (not in NeXus yet...)

# releasing line-too-long restriction to avoid having line breaks in the mapping table
# made the experience that when having a widescreen working with the mapping table
# as single-line instructions is more convenient to read and parsable by human eye


NxImageAngSpaceDict = {"IGNORE": {"fun": "load_from", "terms": "type"},
                       "IGNORE": {"fun": "load_from", "terms": "uuid"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/start_time": {"fun": "convert_iso8601", "terms": ["created", "timezone"]},
                       "IGNORE": {"fun": "load_from", "terms": "is_sequence"},
                       "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/offset"},
                       "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/scale"},
                       "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/units"},
                       "IGNORE": {"fun": "load_from", "terms": "dimensional_calibrations"},
                       "IGNORE": {"fun": "load_from", "terms": "timezone_offset"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/binning": {"fun": "load_from", "terms": "metadata/hardware_source/autostem/Acquisition:Binning"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/dark_mode": {"fun": "load_from", "terms": "metadata/hardware_source/autostem/Acquisition:DarkMode"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/exposure_time": {"fun": "load_from", "terms": "metadata/hardware_source/autostem/Acquisition:ExposureTime(s)"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/exposure_time": "s",
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/gain_mode": {"fun": "load_from", "terms": "metadata/hardware_source/autostem/Acquisition:GainMode"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/flipped": {"fun": "load_from", "terms": "metadata/hardware_source/autostem/Acquisition:IsFlipped"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/readout_tlbr": {"fun": "load_from", "terms": ["metadata/hardware_source/autostem/Acquisition:ReadOutTop", "metadata/hardware_source/autostem/Acquisition:ReadOutLeft", "metadata/hardware_source/autostem/Acquisition:ReadOutBottom", "metadata/hardware_source/autostem/Acquisition:ReadOutRight"]},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/PROCESS[process]/normalization": {"fun": "load_from", "terms": "metadata/hardware_source/autostem/Acquisition:ValueNormalization"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/source"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/timestamp"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/sensor_dimensions_hw"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/sensor_readout_area_tlbr"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/is_flipped_horizontally"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/is_gain_corrected"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/is_dark_subtracted"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/frame_number"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/time_point_ns"},
                       "IGNORE": "ns",
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/integration_count"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/counts_per_electron"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE[ronchicam]/detector_identifier": {"fun": "load_from", "terms": "metadata/hardware_source/hardware_source_id"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/hardware_source_name"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/exposure"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/binning"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/signal_type"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/valid_rows"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/frame_index"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/channel_index"},
                       "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/reference_key"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/voltage": {"fun": "load_from", "terms": "metadata/instrument/high_tension"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/voltage/@units": "V",
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/defocus": {"fun": "load_from", "terms": "metadata/instrument/defocus"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/defocus": "m",
                       "IGNORE": {"fun": "load_from", "terms": "title"},
                       "IGNORE": {"fun": "load_from", "terms": "session_id"},
                       "IGNORE": {"fun": "load_from", "terms": "session"},
                       "IGNORE": {"fun": "load_from", "terms": "category"},
                       "IGNORE": {"fun": "load_from", "terms": "version"},
                       "IGNORE": {"fun": "load_from", "terms": "modified"},
                       "IGNORE": {"fun": "load_from", "terms": "data_shape"},
                       "IGNORE": {"fun": "load_from", "terms": "data_dtype"},
                       "IGNORE": {"fun": "load_from", "terms": "collection_dimension_count"},
                       "IGNORE": {"fun": "load_from", "terms": "datum_dimension_count"},
                       "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/end_time": {"fun": "convert_iso8601", "terms": ["data_modified", "timezone"]},
                       "IGNORE": {"fun": "load_from", "terms": "__large_format"}}



import re

# a custom sphinx extension that is connected to the source-read hook for rst files,
# the purpose is to read all of the contributor information from the rst file and
# place it in a string variable that will be used in the sourcelink.html jinja template
# that has been over ridden and sits in the _templates directory to produce the
# contributor information on the for right sidebar of the html pages

variables_re = re.compile(r"\|(\w+)\| replace::\s(.+)")


def extract_contributor_vars(app, docname, source):
    # Read the RST file content
    content = source[0]

    # Extract variables using regular expressions
    variables = variables_re.findall(content)

    # Create a dictionary to store the extracted variables
    # this will create a list of single strings each of which contain the info about the contributor
    extracted_variables = [var[1] for var in variables]
    if "variables" not in app.config.html_context.keys():
        app.config.html_context["variables"] = {}

    # Add the extracted variables to the Jinja environment
    app.config.html_context["variables"][docname] = extracted_variables


def setup(app):
    app.connect("source-read", extract_contributor_vars)



#!/usr/bin/env python
"""
Writes a simple NeXus HDF5 file using h5py with links
according to the example from Figure 2.1 in the Design chapter
"""

from pathlib import Path

import numpy

from nexusformat.nexus import (NXdata, NXdetector, NXentry, NXfield,
                               NXinstrument, NXlink, nxopen)

filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
buffer = numpy.loadtxt(filename).T
tthData = buffer[0]  # float[]
countsData = numpy.asarray(buffer[1], "int32")  # int[]

with nxopen("simple_example_write2.hdf5", "w") as f:  # create the HDF5 NeXus file
    f["entry"] = NXentry()
    f["entry/instrument"] = NXinstrument()
    f["entry/instrument/detector"] = NXdetector()

    # store the data in the NXdetector group
    f["entry/instrument/detector/two_theta"] = NXfield(tthData, units="degrees")
    f["entry/instrument/detector/counts"] = NXfield(countsData, units="counts")

    f["entry/data"] = NXdata(NXlink("/entry/instrument/detector/counts"),
                             NXlink("/entry/instrument/detector/two_theta"))
    f["entry/data"].set_default()
